#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <regex.h>
#include <libgen.h>
#include <dirent.h>

// libsndfile
#include <sndfile.h>

// STL includes
using namespace std;
#include <string>
#include <vector>
#include <algorithm>
#include <iostream>

class XIBuilder {
private:
	string trackerName;

	/* VALUES FILLED IN BY USER (PARAMETERS) */
	// Name of instrument file.
	string xiName;
	// Name of instrument.
	string instrName;
	// Template name of the instrument samples (could contain a path).
	string sampleName;
	// The maximum number of samples to use.
	int maxSamples;
	int sampleVolume;
	int samplePanning;
	float gain;

	/* VALUES FILLED BY THE PROGRAM */
	// Each member is a key on the keyboard and the number is the associated sample.
	char keySamples[96];
	// A vector of pairs. The first of the pair is a note, denoted as an int starting at C0,
	// and the second is the associated sample.
	vector< pair<int, string> > sampleNotes;
	// Samples represented as an index in sampleNotes. This will never be larger than 16.
	vector<int> samples;

	void buildXI();
	void calcDistribution();
	void parseParams(vector<string> params);

	void xiWrite(const void* data, int items, int itemsize, FILE* fptr);

	string extractNote(string file);

	string pad(string orig, int len);
	string replaceAll(string str, string oldS, string newS);
	int keyNumber(string note);
	short* getSampleData(string &filename, bool getDeltas, int* len);
	template <class T> T* convert2Deltas(T* data, int len, int numChan, int chan);

public:
	XIBuilder();
	int run(vector<string> params);

};

// A note comparison class
class NoteLess {
public:
	bool operator()(pair<int, string> p1, pair<int, string> p2);
} noteLess;
bool NoteLess::operator()(pair<int, string> p1, pair<int, string> p2) {
	if (p1.first < p2.first) return true;
	else return false;
	/*if (p1.first[1] < p2.first[1]) {
		return true;
	} else if (p1.first[1] > p2.first[1]) {
		return false;
	} else {
		if (p1.first[0] >= 'A' && p1.first[0] <= 'B') {
			p1.first[0] += 7;
		}
		if (p2.first[0] >= 'A' && p2.first[0] <= 'B') {
			p2.first[0] += 7;
		}
		if (p1.first[0] < p2.first[0]) {
			return true;
		} else {
			return false;
		}
	}*/
}

string XIBuilder::pad(string orig, int len) {
	int origlen = orig.size();
	
	if (origlen > len) {
		return orig.substr(0, len);
	}

	while (len-- > origlen) {
		orig += " ";
	}
	return orig;
}

XIBuilder::XIBuilder() {
	xiName = "xiinstrument.xi";
	instrName = "XI instrument";
	trackerName = "XIBuilder";
	maxSamples = 16;
	sampleVolume = 64;
	samplePanning = 0;
	gain = 1.0;
}

void XIBuilder::xiWrite(const void* data, int itemsize, int items, FILE* fptr){
	if (fwrite(data, itemsize, items, fptr) != items) {
		if (ferror(fptr)) {
			throw (string)strerror(errno);
		}
		throw (string)"Unknown error";
	}
}

int XIBuilder::run(vector<string> params) {
	try {
		parseParams(params);
		calcDistribution();
		buildXI();
		return 0;
	} catch (string error) {
		fprintf(stderr, "An error occurred: %s\n", error.c_str());
		return 2;
	}
}

void XIBuilder::parseParams(vector<string> params) {
	bool optionsDone = 0;
	bool fileNameDone = 0;
	for (int c = 0; c < params.size(); c++) {
		// Just convenience
		string& p = params[c];

		if (p.size() < 1) continue;

		if (p == "-i" || p == "--instrumentname") {
			if (++c >= params.size()) {
				throw p + " requires an argument";
			}
			instrName = params[c];
		} else if (p == "-o" || p == "--output") {
			if (++c >= params.size()) {
				throw p + " requires an argument";
			}
			xiName = params[c];
		} else if (p == "-n" || p == "--maxsamples") {
			if (++c >= params.size()) {
				throw p + " requires an argument";
			}
			maxSamples = atoi(params[c].c_str());
			if (maxSamples < 1 || maxSamples > 16) {
				throw p + " requires a number between 1 and 16";
			}
		} else if (p == "-v" || p == "--volume") {
			if (++c >= params.size()) {
				throw p + " requires an argument";
			}
			sampleVolume = atoi(params[c].c_str());
			if (sampleVolume < 0 || sampleVolume > 64) {
				throw p + " requires a number between 0 and 64";
			}
		} else if (p == "-p" || p == "--panning") {
			if (++c >= params.size()) {
				throw p + " requires an argument";
			}
			samplePanning = atoi(params[c].c_str());
			if (samplePanning < -128 || samplePanning > 127) {
				throw p + " requires a number between -128 and 127";
			}
		} else if (p == "-g" || p == "--gain") {
			if (++c >= params.size()) {
				throw p + " requires an argument";
			}
			gain = -1.0;
			sscanf(params[c].c_str(), "%f", &gain);
			if (gain < 0 || gain > 1000) {
				throw p + " requires a factor between 0 and 1000";
			}
		} else if (p == "-h" || p == "--help") {
			printf("Usage: xibuilder [options] [--] instrumentfiles\n");
			printf("Format of instrumentfiles: A regular path to a wav file. Replace\n");
			printf("the note in the name with a '%%' (e.g. pingC5.wav -> ping%%.wav).\n");
			printf("The rest of the wave files must follow the same pattern.\n");
			printf("\n");
			printf("Options:\n");
			printf("-h, --help             This summary\n");
			printf("-o, --output           The name of the output file\n");
			printf("                       (default \"%s\")\n", xiName.c_str());
			printf("-i, --instrumentname   The name that will appear in the instrument\n");
			printf("                       (default \"%s\")\n", instrName.c_str());
			printf("-n, --maxsamples       The maximum number of instruments to use\n");
			printf("                       (1-16, default: %i)\n", maxSamples);
			printf("-v, --volume           Sample volume (0-64, default %i)\n", sampleVolume);
			printf("-p, --panning          Sample panning (-128 to 127, default %i)\n", samplePanning);
			printf("-g, --gain             Volume gain (factor, default %.1f)\n", gain);
			exit(0);
		} else if (p == "--") {
			optionsDone = 1;
		} else if (!optionsDone && p[0] == '-') {
			throw p + " is not a valid option";
		} else {
			if (fileNameDone) {
				throw (string)"Only one file name allowed";
			}
			sampleName = params[c];
			fileNameDone = 1;
			if ((signed)sampleName.find("%") < 0) {
				throw (string)"Missing '%' in filename";
			}
		}
	}
}

void XIBuilder::calcDistribution() {
	char dirNameChr[sampleName.size()+1];
	strcpy(dirNameChr, sampleName.c_str());
	string dirName = dirname(dirNameChr);
	
	// The user may not have specified a path, so then we use the current directory.
	if (dirName == "") {
		dirName = ".";
	}

	DIR* dir = opendir(dirName.c_str());
	if (dir == NULL) throw (string)strerror(errno);
	dirent* entry;
	while ((entry = readdir(dir)) != NULL) {
		// Test whether this is one of the files we wish to use.
		string currFile = dirName + "/" + entry->d_name;
		int note = keyNumber(extractNote(currFile));
		if (note < 0) continue;

		// Add the note to the list of available notes.
		pair<int, string> notePair(note, currFile);
		sampleNotes.push_back(notePair);
	}
	closedir(dir);

	if (sampleNotes.size() == 0) {
		throw (string)"Sample files not found";
	}

	// Now sort the notes.
	sort(sampleNotes.begin(), sampleNotes.end(), noteLess);

	// To find the proper distribution of notes (in case there are more than the
	// allowed 16), we use the crude algorithm of dividing the notes evenly across
	// the selection. This might produce bad instruments if there are more notes
	// in one octave than another, but should work reasonably well.
	if (sampleNotes.size() <= maxSamples) {
		for (int c = 0; c < sampleNotes.size(); c++) {
			samples.push_back(c);
		}
	} else {
		// The threshold interval from which we add new samples.
		int inc = sampleNotes.size()-1;
		int distPos = 0;
		int oldDistPos = 1000000;
		for (int c = 0; c < sampleNotes.size(); c++) {
			// Each time we pass over the threshold, add another sample.
			if (distPos < oldDistPos) {
				samples.push_back(c);
			}
			oldDistPos = distPos;
			distPos = (distPos + (maxSamples-1)) % inc;
		}
	}

	// Now we fill the keyboard keys with the samples.
	int keyStart = sampleNotes[samples[0]].first;
	int keyEnd = sampleNotes[samples[samples.size()-1]].first;
	// Fill the initial values.
	for (int c = 0; c < keyStart; c++) {
		keySamples[c] = 0;
	}
	// Fill all the intermediate intervals.
	for (int i = 1; i < samples.size(); i++) {
		int ss, sm, se; // Start, middle, end
		ss = sampleNotes[samples[i-1]].first;
		se = sampleNotes[samples[i]].first;
		sm = (ss+se) / 2;
		for (int j = ss; j <= sm; j++) {
			keySamples[j] = i-1;
		}
		for (int j = sm+1; j < se; j++) {
			keySamples[j] = i;
		}
	}
	// Fill the last values.
	for (int c = keyEnd; c < 96; c++) {
		keySamples[c] = samples.size()-1;
	}
}

int XIBuilder::keyNumber(string note) {
	if (note.size() < 2) return -1;

	int adjust = 0;
	int octi;
	if (note.size() > 2) {
		octi = 2;
		// Check for half tone indications.
		if (note[1] == '#') adjust = 1;
		else if (note[1] == 'b') adjust = -1;
		else return -1;
	} else {
		octi = 1;
	}
	if (note[0] >= 'A' && note[0] <= 'B') {
		note[0] += 7;
	}
	int octave = note[octi] - '0';
	int tone = note[0] - 'C';
	// Here is the tricky part, because we have to skip half tones.	
	int sum = octave * 12 + tone * 2;
	sum -= (tone >= 3) ? 1 : 0; // One possible half tone in the given octave.
	return sum + adjust;
}

string XIBuilder::replaceAll(string str, string oldS, string newS) {
	int last = 0;
	while ((last = str.find(oldS, last)) >= 0) {
		str.replace(last, oldS.size(), newS);
		last += newS.size();
	}
	return str;
}

string XIBuilder::extractNote(string file) {
	string searchStr = sampleName;
	searchStr = replaceAll(searchStr, "{", "\\{");
	searchStr = replaceAll(searchStr, "*", "\\*");
	searchStr = replaceAll(searchStr, "+", "\\+");
	searchStr = replaceAll(searchStr, ".", "\\.");
	searchStr = replaceAll(searchStr, "[", "\\[");
	searchStr = replaceAll(searchStr, "(", "\\(");
	searchStr = replaceAll(searchStr, "%", "([A-H][-#b]{0,1}[0-9])");
	regex_t reg;
	regmatch_t matches[2];

	if (regcomp(&reg, searchStr.c_str(), REG_EXTENDED | REG_ICASE)) {
		throw (string)strerror(errno);
	}

	if (regexec(&reg, file.c_str(), 2, matches, 0)) {
		// No match. Let's bail.
		return "";
	}
	if (matches[1].rm_so < 0) {
		// No match. Let's bail.
		return "";
	}

	string ans(file.substr(matches[1].rm_so, matches[1].rm_eo - matches[1].rm_so));
	if (ans[0] >= 'a' && ans[0] <= 'z') ans[0] += (-'a') + 'A';
	if (ans[1] >= 'A' && ans[1] <= 'Z') ans[1] += (-'A') + 'a';
	ans = replaceAll(ans, "H", "B");
	return ans;
}

void XIBuilder::buildXI() {
	char tmpchr;
	short tmpshrt;
	long tmplong;

	FILE* xiFile = fopen(xiName.c_str(), "wb");
	if (xiFile == NULL) {
		throw (string)"Error opening instrument file: " + strerror(errno);
	}
	
	try {
		// XI header.
		xiWrite("Extended Instrument: ", sizeof(char), 21, xiFile);
		// Instrument name.
		xiWrite(pad(instrName, 22).c_str(), sizeof(char), 22, xiFile);
		// Constant value.
		tmpchr = 0x1a;
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Tracker name.
		xiWrite(pad(trackerName, 20).c_str(), sizeof(char), 20, xiFile);
		// Version number.
		tmpshrt = 0x0102;
		xiWrite(&tmpshrt, sizeof(short), 1, xiFile);

		// Sample distribution over the keyboard.
		xiWrite(keySamples, sizeof(char), 96, xiFile);

		// Twelve volume envelope points with twelve volumes.
		short env[24];
		memset(env, 0, 24*sizeof(short));
		xiWrite(env, sizeof(short), 24, xiFile);
		// Twelve panning envelope points with twelve volumes.
		xiWrite(env, sizeof(short), 24, xiFile);

		// Number of volume points.
		tmpchr = 1;
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Number of panning points.
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Volume sustain point
		tmpchr = 0;
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Volume loop start point
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Volume loop end point
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Panning sustain point
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Panning loop start point
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Panning loop end point
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Volume type;   b0=on, b1=sustain, b2=loop
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Panning type;  b0=on, b1=sustain, b2=loop
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);

		// Vibrato type
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Vibrato sweep
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Vibrato depth
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Vibrato rate
		xiWrite(&tmpchr, sizeof(char), 1, xiFile);
		// Volume fadeout.
		tmpshrt = 0;
		xiWrite(&tmpshrt, sizeof(short), 1, xiFile);
		// Zeroes.
		char zeroes[22];
		memset(zeroes, 0, sizeof(char)*22);
		xiWrite(zeroes, sizeof(char), 22, xiFile);
		
		// Number of samples.
		tmpshrt = samples.size();
		xiWrite(&tmpshrt, sizeof(short), 1, xiFile);

		// Now go through all the samples.
		for (int c = 0; c < samples.size(); c++) {
			int len = 0;
			getSampleData(sampleNotes[samples[c]].second, false, &len);

			// Sample length
			tmplong = len * sizeof(short);
			xiWrite(&tmplong, sizeof(long), 1, xiFile);
			// Sample loop start
			tmplong = 0;
			xiWrite(&tmplong, sizeof(long), 1, xiFile);
			// Sample loop end
			tmplong = 0;
			xiWrite(&tmplong, sizeof(long), 1, xiFile);
			// Volume
			tmpchr = sampleVolume;
			xiWrite(&tmpchr, sizeof(char), 1, xiFile);
			// Fine tune
			tmpchr = 0;
			xiWrite(&tmpchr, sizeof(char), 1, xiFile);
			// Sample type (16 bit)
			tmpchr = 16;
			xiWrite(&tmpchr, sizeof(char), 1, xiFile);
			// Panning
			tmpchr = (samplePanning+128);
			xiWrite(&tmpchr, sizeof(char), 1, xiFile);
			// Relative note (C4 is zero)
			tmpchr = -(sampleNotes[samples[c]].first - 65);
			//tmpchr = 0;
			xiWrite(&tmpchr, sizeof(char), 1, xiFile);
			// Sample name length
			char baseNameBuf[sampleNotes[samples[c]].second.size()];
			strcpy(baseNameBuf, sampleNotes[samples[c]].second.c_str());
			string sampleName = basename(baseNameBuf);
			tmpchr = sampleName.size() < 22 ? sampleName.size() : 22;
			xiWrite(&tmpchr, sizeof(char), 1, xiFile);
			// Sample name
			xiWrite(pad(sampleName, 22).c_str(), sizeof(char), 22, xiFile);
		}
		// Now write the sample data.
		for (int c = 0; c < samples.size(); c++) {
			int len = 0;
			short* deltas = getSampleData(sampleNotes[samples[c]].second, true, &len);
			try {
				// Write the actual sound data.
				xiWrite(deltas, sizeof(short), len, xiFile);
				delete [] deltas;
			} catch (string error) {
				delete [] deltas;
				throw error;
			}
		}
	} catch (string error) {
		fclose(xiFile);
		throw error;
	}

	fclose(xiFile);
}

// getDeltas is there so we can get format data without reading and converting all the samples.
short* XIBuilder::getSampleData(string &filename, bool getDeltas, int* len) {
	SF_INFO sfInfo;
	sfInfo.format = 0;
	SNDFILE* sndFile;
	
	// Open the sound file.
	sndFile = sf_open(filename.c_str(), SFM_READ, &sfInfo);
	*len = sfInfo.frames;

	if (!getDeltas) {
		sf_close(sndFile);
		return NULL;
	}

	int datalen = sfInfo.frames*sfInfo.channels;
	short *data = new short[datalen];
	if (sf_read_short(sndFile, data, datalen) < datalen) {
		delete [] data;
		throw (string)"Error reading soundfile: " + filename + strerror(errno);
	}
	short *deltas = convert2Deltas(data, *len, sfInfo.channels, 0);
	delete [] data;
	sf_close(sndFile);
	return deltas;
}

template <class T> T* XIBuilder::convert2Deltas(T* data, int len, int numChan, int chan) {
	T last = 0;
	T* newdata = new T[len];
	int datalen = len * numChan;
	for (int c = chan, i = 0; c < datalen; c += numChan, i++) {
		// We also appli gain values here.
		T item = (T)((float)data[c] * gain);
		newdata[i] = item - last;		
		last = item;
	}
	return newdata;
}

int main(int argc, char** argv) {
	vector<string> params;

	for (int c = 1; c < argc; c++) {
		params.push_back(argv[c]);
	}

	XIBuilder xiBuilder;

	return xiBuilder.run(params);
}
